unit ModemDetect1;

{  This project demonstrates how to detect all com ports on a machine, then
   detect if a GSM AT SMS modem is connected to one of them. If one is found it
   reports the port where it is connected, the brand, type and software version.

   It demonstrate also on how to use all ports simultanously without using a
   thread, using coorporative multitasking, creating for each port a component
   and execute it at the same time, and how to generate a timeout.

   The TFindATModem also introduce a Release method to make it possible to
   destroy a component from within his own events, using a hidden window and
   posting a message to it.

   You can use or modify this code free for your own needs. }

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, Buttons, ExtCtrls, uCiaComport;

const
   aBaud: array[0..1] of word = (9600, 19200);
   WM_RELEASEPORT = WM_USER + 1;

type
  TModemEvent = (meTry, meBusy, meFound, meBrand, meVersion, meType, meFailed, meTimedOut);
  TFindATModem = class
  private
    FHandle: HWND;
    FPort: TCiaComPort;
    FTimer: TTimer;
    nBaud: integer;
    FClosing: boolean;
    FEvent: TModemEvent;
    FOnEvent: TNotifyEvent;
    FRcvd: PChar;
    procedure WndProc(var MsgRec: TMessage);
    procedure PortDataAvailable(Sender: TObject);
    procedure TimerTimer(Sender: TObject);
    procedure Release;
  public
    constructor Create;
    destructor  Destroy; override;
    procedure Start;
    property Event: TModemEvent read FEvent;
    property Port: TCiaComPort read FPort;
    property Rcvd: PChar read FRcvd;
    property OnEvent: TNotifyEvent read FOnEvent write FOnEvent;
  end;

type
  TMain = class(TForm)
    Panel2: TPanel;
    Memo: TMemo;
    procedure FormActivate(Sender: TObject);
  private
    procedure DetectPorts;
    procedure Log(const ToLog: string);
    procedure TestModemEvent(Sender: TObject);
  end;

var
  Main: TMain;

implementation

{$R *.DFM}

//------------------------------------------------------------------------------
//---- TMain -------------------------------------------------------------------
//------------------------------------------------------------------------------
procedure TMain.Log(const ToLog: string);
begin
   Memo.Lines.Add(TimeToStr(Now) + ' ' + ToLog);
end;

//------------------------------------------------------------------------------
procedure TMain.FormActivate(Sender: TObject);
begin
   Update;
   DetectPorts;
end;

//------------------------------------------------------------------------------
procedure TMain.DetectPorts;
var
   PortList: TStrings;
   n: integer;
begin
   PortList := TStringList.Create;
   try
      CiaEnumPorts(PortList);
      for n := 0 to PortList.Count - 1 do begin
         Log('COM' + PortList[n] + ': port found');

         // create as many components as ports and assign same event to it
         with TFindATModem.Create do begin
            Port.Port := StrToInt(PortList[n]);
            OnEvent   := TestModemEvent;
            Start;
         end;
      end;
   finally
      PortList.Free;
   end;
end;

//------------------------------------------------------------------------------
procedure TMain.TestModemEvent(Sender: TObject);
begin
   // we use sender argument to see from wich component the event is
   with Sender as TFindATModem do
      case Event of
         meTry:      Log('COM' + IntToStr(Port.Port) + ': trying at ' + IntToStr(Port.Baudrate) + ' baud');
         meBusy:     Log('COM' + IntToStr(Port.Port) + ': port is in use');
         meFound:    Log('COM' + IntToStr(Port.Port) + ': modem found');
         meBrand:    Log('COM' + IntToStr(Port.Port) + ': brand:   ' + Rcvd);
         meVersion:  Log('COM' + IntToStr(Port.Port) + ': version: ' + Rcvd);
         meType:     Log('COM' + IntToStr(Port.Port) + ': type:    ' + Rcvd);
         meFailed:   Log('COM' + IntToStr(Port.Port) + ': failed');
         meTimedOut: Log('COM' + IntToStr(Port.Port) + ': timed out');
      end;
end;

//------------------------------------------------------------------------------
//---- TFindATModem ------------------------------------------------------------
//------------------------------------------------------------------------------
constructor TFindATModem.Create;
begin
   FHandle               := AllocateHWND(WndProc);
   FPort                 := TCiaComPort.Create(nil);
   FTimer                := TTimer.Create(nil);
   FPort.Baudrate        := aBaud[nBaud];
   FPort.LineEnd         := 'OK'#13#10;   // answer we need includes the 'OK'
   FPort.LineMode        := True;
   FPort.OnDataAvailable := PortDataAvailable;
end;

//------------------------------------------------------------------------------
procedure TFindATModem.Start;
begin
   try
      FPort.Open := True;
      FEvent     := meTry;
      if Assigned(FOnEvent) then
         FOnEvent(Self);
      FPort.SendStr('AT'#13#10);

      FTimer.Interval := 350;       // 160 mS seems enough for most GSM's
      FTimer.OnTimer  := TimerTimer;
      FTimer.Enabled  := True;
   except
      FEvent := meBusy;             // port already open
      if Assigned(FOnEvent) then
         FOnEvent(Self);
      Release;
   end;
end;

//------------------------------------------------------------------------------
destructor TFindATModem.Destroy;
begin
   FTimer.Destroy;
   FPort.Destroy;
   DeAllocateHWND(FHandle);
   inherited;
end;

//------------------------------------------------------------------------------
procedure TFindATModem.WndProc(var MsgRec: TMessage);
begin
   try
      with MsgRec do
         case Msg of
            WM_RELEASEPORT, WM_QUIT:
               Destroy;
            else
               Result := DefWindowProc(FHandle, Msg, wParam, lParam);
         end;
   except
      // Executed in background, so we have to trap all exceptions here
      // if we do not then program will dissapear in case of exception
   end;
end;

//------------------------------------------------------------------------------
procedure TFindATModem.Release;
begin
   if FClosing then
      Exit;

   FClosing       := True;
   FTimer.Enabled := False;
   PostMessage(FHandle, WM_RELEASEPORT, 0, 0);
end;

//------------------------------------------------------------------------------
procedure TFindATModem.PortDataAvailable(Sender: TObject);
var
   i, j: integer;
   P: PChar;
begin
   j := FPort.RxCount;
   GetMem(P, j);
   try
      // we receive in a PChar because this eliminating of all CrLf and giving
      // the received data to the main application will be much faster
      FPort.Receive(P, j);
      Dec(j);
      Dec(j, Length(FPort.LineEnd));
      while (j > 0) and (P[j] in [#13, #10]) do
         Dec(j);
      i := j;
      Inc(j);
      while (i > 0) and not (P[i] in [#13, #10]) do
         Dec(i);
      if P[i] in [#13, #10] then
         Inc(i);
      P[j] := #0;
      FRcvd := P + i;

      if FEvent = meTry then
         FEvent := meFound;
      if Assigned(FOnEvent) then
         FOnEvent(Self);

      case FEvent of
         meFound: begin
            FEvent := meBrand;
            FPort.SendStr('AT+CGMI'#13#10);
         end;
         meBrand: begin
            FEvent := meVersion;
            FPort.SendStr('AT+CGMR'#13#10);
         end;
         meVersion: begin
            FEvent := meType;
            FPort.SendStr('AT+CGMM'#13#10);
         end;
         else
            Release;
      end;
   finally
      FreeMem(P);
   end;
end;

//------------------------------------------------------------------------------
procedure TFindATModem.TimerTimer(Sender: TObject);
begin
   FTimer.Enabled := False;

   try
      if FEvent <> meTry then begin
         FEvent := meTimedOut;
         Release;
         Exit;
      end;

      inc(nBaud);
      if nBaud > High(aBaud) then begin
         FEvent := meFailed;
         Release;
         Exit;
      end;

      FPort.Baudrate := aBaud[nBaud];
      FPort.SendStr(#13#10'AT'#13#10);    // start with a clean command in the right baudrate
      FTimer.Enabled := True;

   finally
      if Assigned(FOnEvent) then
         FOnEvent(Self);
   end;
end;

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
end.
