unit Thread1;

{  This example shows you how to operate TCiaComPort in a thread. Read carefully
   all comments to not have a false thread.

   The application start a thread and "wait" (by receiving a custom message) for
   the thread to finish.

   The thread send a sequence of commands to the modem, wait for the modem to
   answer with OK and send the next command. When all commands are sent, the
   thread signals the main application which will destroy it. }

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  uCiaComport, StdCtrls, ExtCtrls;

type
   TModemState = (msAT,
                  msATF,
                  msATE,
                  msAT2);

const
   aModemCmd: array[TModemState] of string = ('AT',
                                              'AT&F',
                                              'ATE1',
                                              'AT');
   WM_LOG_FROM_MODEM    = WM_USER + 1;
   WM_MODEM_INITIALIZED = WM_USER + 2;
   WM_MODEMTHREAD_ERROR = WM_USER + 3;
   WM_MODEMTIMEOUT      = WM_USER + 4;

type
  TCommThread = class(TThread)
  private
    FCom: TCiaComPort;
    ModemState: TModemState;
    TimeoutTimer: TTimer;
    procedure ComDataAvailable(Sender: TObject);
    procedure TimeoutTimerTimer(Sender: TObject);
    procedure Error(E: Exception);
  public
    procedure Execute; override;
  end;

type
  TForm1 = class(TForm)
    Panel1: TPanel;
    Memo: TMemo;
    StartBtn: TButton;
    procedure FormCloseQuery(Sender: TObject; var CanClose: Boolean);
    procedure StartBtnClick(Sender: TObject);
  private
    CommThread: TCommThread;
    procedure WMLogFromModem(var Msg: TMessage); message WM_LOG_FROM_MODEM;
    procedure WMModemInitialized(var Msg: TMessage); message WM_MODEM_INITIALIZED;
    procedure WMModemThreadError(var Msg: TMessage); message WM_MODEMTHREAD_ERROR;
    procedure WMModemTimeout(var Msg: TMessage); message WM_MODEMTIMEOUT;
  public
  end;

var
  Form1: TForm1;

implementation

{$R *.DFM}

//------------------------------------------------------------------------------
//---- Main thread code --------------------------------------------------------
//------------------------------------------------------------------------------
procedure TForm1.StartBtnClick(Sender: TObject);
begin
   CommThread := TCommThread.Create(False);
   StartBtn.Enabled := False;
end;

//------------------------------------------------------------------------------
procedure TForm1.FormCloseQuery(Sender: TObject; var CanClose: Boolean);
begin
   if Assigned(CommThread) then begin
      PostThreadMessage(CommThread.ThreadID, WM_QUIT, 0, 0);
      CommThread.WaitFor;
   end;
end;

//------------------------------------------------------------------------------
procedure TForm1.WMLogFromModem(var Msg: TMessage);
var
   P: PChar;
   C: string;
begin
   P := PChar(Msg.WParam);
   C := Trim(P);
   if C > '' then
      Memo.Lines.Add('> ' + C);
   Dispose(P);
end;

//------------------------------------------------------------------------------
procedure TForm1.WMModemInitialized(var Msg: TMessage);
begin
   Memo.Lines.Add('Modem initialized');
   PostThreadMessage(CommThread.ThreadID, WM_QUIT, 0, 0);
   CommThread.WaitFor;
   CommThread := nil;
   StartBtn.Enabled := True;
end;

//------------------------------------------------------------------------------
procedure TForm1.WMModemTimeout(var Msg: TMessage);
begin
   Memo.Lines.Add('Modem timed out');
   PostThreadMessage(CommThread.ThreadID, WM_QUIT, 0, 0);
   CommThread.WaitFor;
   CommThread := nil;
   StartBtn.Enabled := True;
end;

//------------------------------------------------------------------------------
procedure TForm1.WMModemThreadError(var Msg: TMessage);
var
   P: PChar;
begin
   Beep;
   P := PChar(Msg.WParam);
   Memo.Lines.Add('Error: ' + P);
   Dispose(P);
   CommThread := nil;
   StartBtn.Enabled := True;
end;

//------------------------------------------------------------------------------
//---- TCommThread -------------------------------------------------------------
//------------------------------------------------------------------------------
procedure TCommThread.Error(E: Exception);
var
   Err: PChar;
begin
   GetMem(Err, Length(E.Message) + 1);
   Move(E.Message[1], Err[0], Length(E.Message) + 1);
   PostMessage(Form1.Handle, WM_MODEMTHREAD_ERROR, integer(Err), 0);
end;

//------------------------------------------------------------------------------
procedure TCommThread.TimeoutTimerTimer(Sender: TObject);
begin
   PostMessage(Form1.Handle, WM_MODEMTIMEOUT, 0, 0);
end;

//------------------------------------------------------------------------------
procedure TCommThread.ComDataAvailable(Sender: TObject);
var
   Rcvd: PChar;
begin
   GetMem(Rcvd, FCom.RxCount + 1);
   Rcvd[FCom.Receive(Rcvd, FCom.RxCount)] := #0;
   PostMessage(Form1.Handle, WM_LOG_FROM_MODEM, integer(Rcvd), 0);

   if Rcvd = 'OK'#13#10 then begin
      Inc(ModemState);
      if ModemState > High(TModemState) then begin
         PostMessage(Form1.Handle, WM_MODEM_INITIALIZED, 0, 0);
         Exit;
      end;
      Sleep(100); // some modems do not like fast sending
      TimeoutTimer.Enabled := False;
      TimeoutTimer.Enabled := True;
      FCom.SendStr(aModemCmd[ModemState] + #13#10);
   end;
end;

//------------------------------------------------------------------------------
procedure TCommThread.Execute;
begin
   try
      {  It is absolutely nececary that the ComPort is created in the thread,
         otherwise you have a false thread and all code will execute in main
         thread context while you only "think" you have threaded code. Note
         that it has to be destroyed in the same thread as where created. Note
         that the only threaded code in a thread is the Execute method, and all
         what is called from there and of course all events from components that
         have a hidden window, like TCiaComPort. }

      FCom                 := TCiaComPort.Create(nil);
      FCom.OnDataAvailable := ComDataAvailable;
      FCom.Port            := 1;
      FCom.LineEnd         := #13#10;
      FCom.LineMode        := True;
      FCom.Open            := True;
      FCom.SendStr(aModemCmd[ModemState] + #13#10);

      TimeoutTimer          := TTimer.Create(nil);
      TimeoutTimer.OnTimer  := TimeoutTimerTimer;
      TimeoutTimer.Interval := 5000;
      TimeoutTimer.Enabled  := True;

      try

         {  We have not mutch to do here, because the CommPort is event driven.
         This loop will exit as soon a WM_QUIT message arrive. }

         FCom.MessageLoop;

      finally
         FCom.Free;
         TimeoutTimer.Free;
      end;

   except
      on E: Exception do
         Error(E);
   end;
end;

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
end.
