unit uBuffer;

{
   TBuffer
      A buffer that in the begin will grow to needed size, and then will re-use
      his allocated memory over and over again to gain speed and to never have
      fragmented memory.

      Methods and properties:
         Clear
            Clear the buffer. Internal it only reset some pointers.

         Write
            Copy data to the buffer and grow when needed

         Read
            Copy data from the buffer and Clear it when empty. Do not use if no
            extra copy is needed, if no extra copy needed use Memory instead.

         Memory
            Pointer to the memory area of the buffer. You can use this if no
            extra copy of the data is needed. If data is not nececary after it
            then call Clear after it.

         Size
            Returns the size of the Data.

   TDebug
      A buffer derived from TBuffer. It is to have data printed in ascii-hex
      debug format. Memory as well as Read will return null terminated data.
      Note also that Size returns the size of the data including the null
      terminating byte.

      Methods and properties:
         DebugStr
            Convert data into ascii hex format before copy it to the data area,
            use Memory or Read to get it back. After calling this method all
            characters in the data with valus less than ' ' are replaced by a
            '.' unless PreserveData is set to True, but this will make an
            additional copy of the data wich is slower of course.

         PreserveData
            If set to True then the Data given as argument to DebugStr will have
            and additional copy and will be preserved. Of course DebugStr will
            take more CPU and thus slower.

   TDSBuffer
      A buffer derived from TBuffer. It has some additional methods and
      properties to make it useable for a DataScoop.

      Methods and properties:
         Header
            Returns a header with datetime information and to show if it is data
            from DTE or from DCE

         IsDCE
            Set the Header according to this setting

   TTimedBuffer
      A buffer derived from TDSBuffer. It has some additional methods and
      properties to make a timeout. Windows does not deliver serial port data
      in real time. In fact it comes in chuncks of typical 8 bytes or with a
      timeout of typical 1 timeslice (10 mSec in W2K). It use a TTimer with
      interval set to 1 (which convert to a timeslice) to deliver his data.

      Methods and properties:
         Timeout
            Normally you can let it the default value of 1. Could be in special
            occasions you have to set the value higher, but most likely it is
            never needed.

         OnDataEnd
            Is fired if serial port that feed the buffer did not received data
            during 1 windows timeslice (typical 10 mSec).
}

interface

uses
   Classes, ExtCtrls, SysUtils;

type
  TBuffer = class
  private
    FData: PChar;
    FReadPtr: integer;
    FWritePtr: integer;
    FCapacity: integer;
    procedure Grow(NewSize: integer);
    function  GetSize: integer;
    function  GetMemory: PChar;
  public
    constructor Create; virtual;
    destructor  Destroy; override;
    procedure Clear;
    procedure Write(const Data; Len: integer); virtual;
    function  Read(var Data; Len: integer): integer;
    property Memory: PChar read GetMemory;
    property Size: integer read GetSize;
  end;

  TDebug = class(TBuffer)
  private
    FPreserveData: boolean;
    function Byte2Hex(c: char): word;
    procedure InternalDebugStr(var Data; Len: integer);
  public
    procedure DebugStr(var Data; Len: integer);
    property PreserveData: boolean read FPreserveData write FPreserveData;
  end;

  TDSBuffer = class(TBuffer)
  private
    FIsDCE: boolean;
    FHeader: string;
    function  GetHeader: string;
    procedure SetIsDCE(const Value: boolean);
  public
    constructor Create; override;
    property Header: string read GetHeader;
    property IsDCE: boolean read FIsDCE write SetIsDCE;
  end;

  TTimedBuffer = class(TDSBuffer)
  private
    FTimer: TTimer;
    FTimeout: integer;
    FOnDataEnd: TNotifyEvent;
    procedure FTimerTimer(Sender: TObject);
    procedure SetTimeout(const Value: integer);
  public
    constructor Create; override;
    destructor  Destroy; override;
    procedure Write(const Data; Len: integer); override;
    property Timeout: integer read FTimeout write SetTimeout;
    property OnDataEnd: TNotifyEvent read FOnDataEnd write FOnDataEnd;
  end;

implementation

{ TBuffer }

const
   INITIAL_SIZE = 128;

procedure TBuffer.Clear;
begin
   FReadPtr  := 0;
   FWritePtr := 0;
end;

constructor TBuffer.Create;
begin
   inherited;
   Grow(INITIAL_SIZE);
end;

destructor TBuffer.Destroy;
begin
   if Assigned(FData) then
      FreeMem(FData);
   inherited;
end;

function TBuffer.GetMemory: PChar;
begin
   Result := FData + FReadPtr;
end;

function TBuffer.GetSize: integer;
begin
   Result := FWritePtr - FReadPtr;
end;

procedure TBuffer.Grow(NewSize: integer);
begin
   if NewSize <= FCapacity then
      Exit;
   FCapacity := NewSize;
   ReallocMem(FData, FCapacity);
end;

function TBuffer.Read(var Data; Len: integer): integer;
begin
   if Size < Len then
      Len := Size;
   Result := Len;
   if Result <= 0 then
      Exit;
   Move(FData[FReadPtr], Data, Len);
   Inc(FReadPtr, Len);
   if FReadPtr >= FWritePtr then
      Clear;
end;

procedure TBuffer.Write(const Data; Len: integer);
var
   P: PChar;
begin
   P := @Data;
   if FCapacity - FWritePtr < Len then
      Grow(FCapacity + Len);
   Move(P[0], FData[FWritePtr], Len);
   Inc(FWritePtr, Len);
end;

{ TDSBuffer }

constructor TDSBuffer.Create;
begin
   inherited;
   FHeader := 'DTE ' + FormatDateTime('dd/mm hh:nn:ss.zzz', Now);
end;

function TDSBuffer.GetHeader: string;
begin
   Move(FormatDateTime('dd/mm hh:nn:ss.zzz', Now)[1], FHeader[5], 18);
   Result := FHeader;
end;

procedure TDSBuffer.SetIsDCE(const Value: boolean);
begin
   if FIsDCE = Value then
      Exit;
   FIsDCE := Value;
   if Value then
      FHeader[2] := 'C'
   else
      FHeader[2] := 'T';
end;

{ TTimedBuffer }

constructor TTimedBuffer.Create;
begin
   inherited;
   FTimeout        := 1;
   FTimer          := TTimer.Create(nil);
   FTimer.Interval := FTimeout;
   FTimer.Enabled  := False;
   FTimer.OnTimer  := FTimerTimer;
end;

destructor TTimedBuffer.Destroy;
begin
   if Assigned(FTimer) then
      FTimer.Destroy;
   inherited;
end;

procedure TTimedBuffer.FTimerTimer(Sender: TObject);
begin
   FTimer.Enabled := False;
   if Size <= 0 then
      Exit;
   if Assigned(FOnDataEnd) then
      FOnDataEnd(Self);
end;

procedure TTimedBuffer.SetTimeout(const Value: integer);
begin
   if FTimeout = Value then
      Exit;
   FTimeout        := Value;
   FTimer.Interval := Value;
end;

procedure TTimedBuffer.Write(const Data; Len: integer);
begin
   inherited;
   FTimer.Enabled := False;
   FTimer.Enabled := True;
end;

{ TDebug }

{function TDebug.Byte2Hex(c: char): word;
type
   TCnvrtWord = packed record
      case byte of
         1: ( Lo: byte;
              Hi: byte;  );
         2: ( Ret: word; );
   end;
var
   Cnvrt: TCnvrtWord;
begin
   Cnvrt.Hi := (byte(c) and $0F) + $30;
   if Cnvrt.Hi > $39 then
      Inc(Cnvrt.Hi, 7);
   Cnvrt.Lo := (byte(c) shr 4) + $30;
   if Cnvrt.Lo > $39 then
      Inc(Cnvrt.Lo, 7);
   Result := Cnvrt.Ret;
end; }

function TDebug.Byte2Hex(c: char): word;
asm
   {  In fact this routine returns a string of 2 characters. We use this trick
      to have a very fast routine without allocating any memory. Some CBuilder
      IDE's dont compile asm code, use Delphi code above in that case. Note that
      you can do exacly the same with a lookup table, But I'm not sure it will
      be faster. Once a while we have to use asm to not forget it anyway :)

      In a class method, the first argument (in eax) is a pointer to Self.
      The 3 first arguments are passed in the registers eax, edx and ecx. Next
      arguments are pushed on the stack. Result value is always in eax. }

         mov al, dl     // get byte
         and al, 0Fh    // filter low order nibble
         add al, 30h    // 0 -> 30h, 1 -> 31h, etc
         cmp al, 39h    // check if A..F
         jbe @2
         add al, 7      // A..F
   @2:   mov ah, al
         mov al, dl     // get byte again
         shr al, 4      // filter high order nibble
         add al, 30h    // 0 -> 30h, 1 -> 31h, etc
         cmp al, 39h    // check if A..F
         jbe @1
         add al, 7      // A..F
   @1:
end;

procedure TDebug.DebugStr(var Data; Len: integer);
var
   CpyData: string;
begin
   if not FPreserveData then begin
      InternalDebugStr(Data, Len);
      Exit;
   end;

   SetLength(CpyData, Len);
   Move(Data, CpyData[1], Len);
   InternalDebugStr(CpyData[1], Len);
end;

procedure TDebug.InternalDebugStr(var Data; Len: integer);
var
   Size, i, j: integer;
   P: PChar;
   Hex: word;
begin
   Size := Len * 3 + Len + Len div 16 * 2;
   if Len mod 16 > 0 then
      Inc(Size, 66);
   if Size > FCapacity then
      Grow(Size);
   P := @Data;
   i := 0;
   while True do begin
      j := 0;
      while (j < 16) and (i < Len) do begin
         Hex := Byte2Hex(P[j]);
         Write(Hex, 2);
         if j = 7 then
            Write('- '[1], 1)
         else
            Write('  '[1], 1);
         if P[j] < ' ' then
            P[j] := '.';      // Data is garbage after this routine
         Inc(j);
         Inc(i);
      end;
      if j < 16 then begin
         //    '00 01 02 03 04 05 06 07-08 09 0A 0B 0C 0D 0E 0F '
         Write('                                                '[1], (16 - j) * 3);
      end;
      Write(P[0], j);
      if i >= Len then begin
         Write(#0#0[1], 1);
         Exit;
      end;
      Inc(P, j);
      Write(#13#10[1], 2);
   end;
end;

end.
